/*
 * Copyright (C) 2012  Radek Novacek <rnovacek@redhat.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include "connection.h"

#include <netinet/in.h>
#include <string.h>

#include "dbus_wrapper.h"
#include "errors.h"

#include "network_nm.h"
#include "nm_support.h"
#include "setting_nm.h"

#include "connection_private.h"
#include "port_private.h"
#include "network_private.h"
#include "setting_private.h"


typedef struct ConnectionPriv {
    DBusGProxy *proxy;
    GHashTable *properties;
    Network *network; // Needed for ability to update connection on Updated signal
} ConnectionPriv;

void connection_get_properties(Network *network, Connection *connection);
void connection_updated_cb(void *proxy, Connection *connection);
void connection_removed_cb(void *proxy, Connection *connection);

Connection *connection_new_from_objectpath(Network *network, const char *objectpath)
{
    // Id and name will be filled later
    Connection *connection = connection_new(NULL, NULL);
    ConnectionPriv *priv = malloc(sizeof(ConnectionPriv));
    connection->priv = priv;
    connection->uuid = strdup(objectpath);
    priv->network = network;
    priv->proxy = dbus_g_proxy_new_for_name(network_priv_get_dbus_connection(network), NM_SERVICE_DBUS, objectpath, NM_DBUS_IFACE_SETTINGS_CONNECTION);
    priv->properties = NULL;

    connection_get_properties(network, connection);

    dbus_g_proxy_add_signal(priv->proxy, "Updated", G_TYPE_INVALID);
    dbus_g_proxy_connect_signal(priv->proxy, "Updated", G_CALLBACK(connection_updated_cb), connection, NULL);

    dbus_g_proxy_add_signal(priv->proxy, "Removed", G_TYPE_INVALID);
    dbus_g_proxy_connect_signal(priv->proxy, "Removed", G_CALLBACK(connection_removed_cb), connection, NULL);
    return connection;
}

void connection_read_properties(Network *network, Connection *connection)
{
    ConnectionPriv *priv = connection->priv;
    GHashTable *connection_map = dbus_property_map(priv->properties, "connection");
    if (!connection_map) {
        warn("Connection %s is invalid", connection->uuid);
        return;
    }
    connection->id = strdup(dbus_property_string(connection_map, "uuid"));
    connection->name = strdup(dbus_property_string(connection_map, "id"));

    GValue *v = g_hash_table_lookup(connection_map, "autoconnect");
    if (v) {
        connection->autoconnect = g_value_get_boolean(v);
    } else {
        connection->autoconnect = false;
    }

    const char *type = dbus_property_string(connection_map, "type");
    if (strcmp(type, "802-3-ethernet") == 0) {
        // TODO: handle other types
        GHashTable *ethernet_map = dbus_property_map(priv->properties, "802-3-ethernet");
        v = g_hash_table_lookup(ethernet_map, "mac-address");
        if (v != NULL) {
            GByteArray *mac = g_value_get_boxed(v);
            char *mac_str = macFromGByteArray(mac);
            connection->port = network_port_by_mac(network, mac_str);
            free(mac_str);
        }
    }

    connection->settings = settings_new(2);
    GHashTable *ip4_map = dbus_property_map(priv->properties, "ipv4");
    GHashTable *ip6_map = dbus_property_map(priv->properties, "ipv6");

    const char *method;
    Setting *setting;
    unsigned int i;
    guint32 prefix;

    if (ip4_map) {
        guint32 address, gateway;
        method = dbus_property_string(ip4_map, "method");
        if (strcmp(method, "auto") == 0) {
            // DHCP
            setting = setting_new_DHCP(connection, IPv4);
            connection_add_setting(connection, setting);
        } else if (strcmp(method, "manual") == 0) {
            // Static address
            GPtrArray *addresses = dbus_property_array(ip4_map, "addresses");
            GArray *array;
            for (i = 0; addresses && i < addresses->len; ++i) {
                array = (GArray *) g_ptr_array_index(addresses, i);
                if (!ipv4_address_from_array(array, &address, &prefix, &gateway)) {
                    warn("IPv4 address (%s) invalid", connection->id);
                    continue;
                }
                setting = setting_new_from_static_ipv4(connection, address, prefix, gateway, i);
                connection_add_setting(connection, setting);
            }
        } else if (strcmp(method, "link-local") == 0) {
            setting = setting_new_link_local(connection, IPv4);
            connection_add_setting(connection, setting);
        } else if (strcmp(method, "disabled") == 0) {
            debug("IPv4 connection is disabled for connection: %s", connection->name);
        } else {
            warn("Unknown IPv4 method: %s", method);
        }

        if (setting != NULL) {
            char *dns;
            GArray *dnss = (GArray *) dbus_property_array(ip4_map, "dns");
            for (i = 0; dnss && i < dnss->len; ++i) {
                dns = ip4ToString(g_array_index(dnss, guint32, i));
                setting_add_dns_server(setting, dns);
            }

            if ((v = g_hash_table_lookup(ip4_map, "dns-search")) != NULL) {
                for (char **dnssearchs = g_value_get_boxed(v);
                     *dnssearchs != NULL;
                     dnssearchs++) {

                    setting_add_search_domain(setting, *dnssearchs);
                }
            }
        }
    }
    if (ip6_map) {
        method = dbus_property_string(ip6_map, "method");
        if (strcmp(method, "auto") == 0) {
            // DHCP
            setting = setting_new_DHCP(connection, IPv6);
            connection_add_setting(connection, setting);
        } else if (strcmp(method, "manual") == 0) {
            // Static or link-local addresses
            GPtrArray *addresses = dbus_property_array(ip6_map, "addresses");
            GByteArray *address, *gateway;
            GValueArray *array;
            for (i = 0; addresses && i < addresses->len; ++i) {
                array = (GValueArray *) g_ptr_array_index(addresses, i);
                if (!ipv6_address_from_array(array, &address, &prefix, &gateway)) {
                    warn("IPv6 address (%s) invalid", connection->id);
                    continue;
                }
                setting = setting_new_from_static_ipv6(connection, address, prefix, gateway, i);
                connection_add_setting(connection, setting);
            }
        } else if (strcmp(method, "link-local") == 0) {
            // Link-local address
            setting = setting_new_link_local(connection, IPv6);
            connection_add_setting(connection, setting);
        } else if (strcmp(method, "ignore") == 0) {
            debug("IPv6 connection is disabled for connection: %s", connection->name);
        } else {
            warn("Unknown IPv6 method: %s", method);
        }

        if (setting != NULL) {
            char *dns;
            GPtrArray *dnss = dbus_property_array(ip6_map, "dns");
            for (i = 0; dnss && i < dnss->len; ++i) {
                dns = ip6ToString(g_ptr_array_index(dnss, i));
                setting_add_dns_server(setting, dns);
            }

            if ((v = g_hash_table_lookup(ip6_map, "dns-search")) != NULL) {
                for (char **dnssearchs = g_value_get_boxed(v);
                     *dnssearchs != NULL;
                     dnssearchs++) {

                    setting_add_search_domain(setting, *dnssearchs);
                }
            }
        }
    }
}

void g_hash_table_insert_string(GHashTable *hash, const char *key, const char *value)
{
    GValue *v = g_slice_new0(GValue);
    g_value_init(v, G_TYPE_STRING);
    g_value_set_string(v, strdup(value));
    g_hash_table_insert(hash, strdup(key), v);
}

void g_hash_table_insert_boxed(GHashTable *hash, const char *key, GType t, gconstpointer value)
{
    GValue *v = g_slice_new0(GValue);
    g_value_init(v, t);
    g_value_set_boxed(v, value);
    g_hash_table_insert(hash, strdup(key), v);
}

GByteArray *ip6ArrayFromString(const char *addr) {
    if (addr != NULL) {
        struct in6_addr *ip = ip6FromString(addr);
        return g_byte_array_new_take(ip->s6_addr, 16);
    } else {
        return g_byte_array_new_take(calloc(sizeof(guint8), 16), 16);
    }
}

GHashTable *connection_to_hash(Connection *connection)
{
    GHashTable *hash = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, (GDestroyNotify) g_hash_table_destroy);
    GHashTable *subhash;
    GHashTable *ipv4hash = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, (GDestroyNotify) g_value_unset);
    GHashTable *ipv6hash = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, (GDestroyNotify) g_value_unset);
    GPtrArray *ip4addresses = g_ptr_array_new();
    GPtrArray *ip6addresses = g_ptr_array_new();
    GArray *ip4addr;
    GValueArray *ip6addr;
    Setting *setting;

    // connection
    subhash = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, (GDestroyNotify) g_value_unset);
    g_hash_table_insert_string(subhash, "id", connection->name);

    if (connection->id == NULL) {
        connection->id = uuid_gen();
    }
    g_hash_table_insert_string(subhash, "uuid", connection->id);

    // TODO: handle other types
    g_hash_table_insert_string(subhash, "type", "802-3-ethernet");

    g_hash_table_insert(hash, strdup("connection"), subhash);

    // connection type specific
    subhash = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, (GDestroyNotify) g_value_unset);
    if (connection->port != NULL) {
        g_hash_table_insert_boxed(subhash, "mac-address", DBUS_TYPE_G_UCHAR_ARRAY, macToGByteArray(connection->port->mac));
    }
    g_hash_table_insert(hash, strdup("802-3-ethernet"), subhash);

    for (unsigned int i = 0; i < settings_length(connection->settings); ++i) {
        setting = settings_index(connection->settings, i);
        switch (setting->type) {
            case SETTING_DHCP:
                if ((setting->protocol_type & IPv4) > 0) {
                    g_hash_table_insert_string(ipv4hash, "method", "auto");
                }
                if ((setting->protocol_type & IPv6) > 0) {
                    g_hash_table_insert_string(ipv6hash, "method", "auto");
                }
                break;
            case SETTING_LINK_LOCAL:
                if ((setting->protocol_type & IPv4) > 0) {
                    g_hash_table_insert_string(ipv4hash, "method", "link-local");
                }
                if ((setting->protocol_type & IPv6) > 0) {
                    g_hash_table_insert_string(ipv6hash, "method", "link-local");
                }
                break;
            case SETTING_STATIC:
                if ((setting->protocol_type & IPv4) > 0) {
                    debug("Gateway: %s", setting->default_gateway);
                    ip4addr = ipv4_address_to_array(
                            ip4FromString(setting->address),
                            setting->prefix,
                            setting->default_gateway == NULL ? 0 : ip4FromString(setting->default_gateway));
                    g_ptr_array_add(ip4addresses, ip4addr);
                    if (!g_hash_table_contains(ipv4hash, "method")) {
                        g_hash_table_insert_string(ipv4hash, "method", "manual");
                    }
                    if (setting->dns_servers != NULL) {
                        GArray *dns = g_array_new(false, true, sizeof(guint32));
                        for (unsigned j = 0; j < dns_servers_length(setting->dns_servers); ++j) {
                            guint32 ip = ip4FromString(dns_servers_index(setting->dns_servers, j));
                            g_array_append_val(dns, ip);
                        }
                        g_hash_table_insert_boxed(ipv4hash, "dns", DBUS_TYPE_G_ARRAY_OF_UINT, dns);
                    }
                    if (setting->search_domains != NULL) {
                        GSList *search = NULL;
                        for (unsigned j = 0; j < search_domains_length(setting->search_domains); ++j) {
                            search = g_slist_append(search, strdup(search_domains_index(setting->search_domains, j)));
                        }
                        g_hash_table_insert_boxed(ipv4hash, "dns-search", DBUS_TYPE_G_LIST_OF_STRING, search);
                    }
                }
                if ((setting->protocol_type & IPv6) > 0) {
                    ip6addr = ipv6_address_to_array(
                            ip6ArrayFromString(setting->address),
                            setting->prefix,
                            ip6ArrayFromString(setting->default_gateway));
                    g_ptr_array_add(ip6addresses, ip6addr);
                    if (!g_hash_table_contains(ipv6hash, "method")) {
                        g_hash_table_insert_string(ipv6hash, "method", "manual");
                    }

                    if (setting->dns_servers != NULL) {
                        GPtrArray *dns = g_ptr_array_new();
                        for (unsigned j = 0; j < dns_servers_length(setting->dns_servers); ++j) {
                            GByteArray *ip = ip6ArrayFromString(dns_servers_index(setting->dns_servers, j));
                            g_ptr_array_add(dns, ip);
                        }
                        g_hash_table_insert_boxed(ipv6hash, "dns", DBUS_TYPE_G_ARRAY_OF_ARRAY_OF_UCHAR, dns);
                    }
                    if (setting->search_domains != NULL) {
                        GSList *search = NULL;
                        for (unsigned j = 0; j < search_domains_length(setting->search_domains); ++j) {
                            search = g_slist_append(search, strdup(search_domains_index(setting->search_domains, j)));
                        }
                        g_hash_table_insert_boxed(ipv6hash, "dns-search", DBUS_TYPE_G_LIST_OF_STRING, search);
                    }
                }
                break;
            default:
                error("Can't create setting with unknown type");
        }
    }
    if (g_hash_table_size(ipv4hash) > 0) {
        if (ip4addresses->len > 0) {
            g_hash_table_insert_boxed(ipv4hash, "addresses", DBUS_TYPE_G_ARRAY_OF_ARRAY_OF_UINT, ip4addresses);
        }
        g_hash_table_insert(hash, strdup("ipv4"), ipv4hash);
    }
    if (g_hash_table_size(ipv6hash) > 0) {
        if (ip6addresses->len > 0) {
            g_hash_table_insert_boxed(ipv6hash, "addresses", DBUS_TYPE_G_ARRAY_OF_IP6_ADDRESS, ip6addresses);
        }
        g_hash_table_insert(hash, strdup("ipv6"), ipv6hash);
    }
    return hash;
}

void connection_priv_free(void *priv)
{
    ConnectionPriv *p = priv;
    if (p->properties != NULL) {
        g_hash_table_destroy(p->properties);
    }
    g_object_unref(p->proxy);
    free(p);
}

void connection_get_properties(Network *network, Connection *connection)
{
    ConnectionPriv *priv = connection->priv;
    if (priv->properties != NULL) {
        g_hash_table_destroy(priv->properties);
    }

    GHashTable *hash;
    GError *err = NULL;
    if (!dbus_g_proxy_call(priv->proxy, "GetSettings", &err, G_TYPE_INVALID,
        DBUS_TYPE_G_MAP_OF_MAP_OF_VARIANT, &hash, G_TYPE_INVALID)) {

        error("Call GetSetting of %s failed: %s", NM_DBUS_IFACE_SETTINGS_CONNECTION, err->message);
        return;
    }
    if (hash == NULL) {
        error("Connection %s doesn't have any settings", connection->uuid);
        return;
    }
    check_connection_hash(hash);
    priv->properties = hash;
    connection_read_properties(network, connection);
}

void connection_updated_cb(void *proxy, Connection *connection)
{
    debug("Connection updated: %s (%s)", connection_get_name(connection), connection_get_uuid(connection));
    ConnectionPriv *priv = connection->priv;
    MUTEX_LOCK(priv->network);
    connection_get_properties(priv->network, connection);
    MUTEX_UNLOCK(priv->network);
}

void connection_removed_cb(void *proxy, Connection *connection)
{
    debug("Connection deleted: %s (%s)", connection_get_name(connection), connection_get_uuid(connection));

    ConnectionPriv *priv = connection->priv;
    // NetworkManager sometimes sends Removed signal twice, we need to workaround it
    if (connection == NULL || strcmp(dbus_g_proxy_get_path(proxy), connection->uuid) != 0) {
        debug("Connection already deleted");
        return;
    }

    MUTEX_LOCK(priv->network);
    unsigned int i = 0;
    for (i = 0; i < connections_length(priv->network->connections); ++i) {
        if (strcmp(connections_index(priv->network->connections, i)->uuid, connection->uuid) == 0) {
            break;
        }
    }
    if (i < connections_length(priv->network->connections)) {
        connection_free(connections_pop(priv->network->connections, i));
        connection = NULL;
    }
    MUTEX_UNLOCK(priv->network);
}

int connection_priv_delete(const Connection *connection)
{
    ConnectionPriv *priv = connection->priv;
    MUTEX_LOCK(priv->network);
    GError *err = NULL;
    if (!dbus_g_proxy_call(priv->proxy, "Delete", &err, G_TYPE_INVALID, G_TYPE_INVALID)) {
        error("Deleting of connection failed: %s", err->message);
        return LMI_ERROR_CONNECTION_DELETE_FAILED;
    }
    MUTEX_UNLOCK(priv->network);
    return LMI_SUCCESS;
}
